@@ -93,6 +93,8 @@ static int validate_checksums(zckCtx *zck, zck_log_type bad_checksums) {
     for(zckChunk *idx = zck->index.first; idx; idx = idx->next) {
         if(idx == zck->index.first && idx->length == 0) {
             idx->valid = 1;
+            if(zck->header_only)
+                break;
             continue;
         }
 
@@ -108,8 +110,10 @@ static int validate_checksums(zckCtx *zck, zck_log_type bad_checksums) {
                 zck_log(ZCK_LOG_DEBUG, "No more data");
             if(!hash_update(zck, &(zck->check_chunk_hash), buf, rsize))
                 return 0;
-            if(!hash_update(zck, &(zck->check_full_hash), buf, rsize))
-                return 0;
+            if(!zck->has_uncompressed_source) {
+                if(!hash_update(zck, &(zck->check_full_hash), buf, rsize))
+                    return 0;
+            }
             rlen += rsize;
         }
         int valid_chunk = validate_chunk(idx, bad_checksums);
@@ -118,18 +122,28 @@ static int validate_checksums(zckCtx *zck, zck_log_type bad_checksums) {
         idx->valid = valid_chunk;
         if(all_good && valid_chunk != 1)
             all_good = false;
+        if(zck->header_only)
+            break;
     }
     int valid_file = -1;
-    if(all_good) {
-        /* Check data checksum */
-        valid_file = validate_file(zck, bad_checksums);
-        if(!valid_file)
-            return 0;
+    if(zck->has_uncompressed_source || zck->header_only) {
+        /* If we have an uncompressed source or are a detached header,
+         * skip meaningless full data checksum, and just set valid_file
+         * if the chunks (or dictionary, if we're a header) was good */
+        if(all_good)
+            valid_file = 1;
+    } else {
+        if(all_good) {
+            /* Check data checksum */
+            valid_file = validate_file(zck, bad_checksums);
+            if(!valid_file)
+                return 0;
 
-        /* If data checksum failed, invalidate *all* chunks */
-        if(valid_file == -1)
-            for(zckChunk *idx = zck->index.first; idx; idx = idx->next)
-                idx->valid = -1;
+            /* If data checksum failed, invalidate *all* chunks */
+            if(valid_file == -1)
+                for(zckChunk *idx = zck->index.first; idx; idx = idx->next)
+                    idx->valid = -1;
+        }
     }
 
     /* Go back to beginning of data section */
@@ -146,6 +160,10 @@ static int validate_checksums(zckCtx *zck, zck_log_type bad_checksums) {
 char *get_digest_string(const char *digest, int size) {
     char *str = zmalloc(size*2+1);
 
+    if (!str) {
+       zck_log(ZCK_LOG_ERROR, "OOM in %s", __func__);
+       return NULL;
+    }
     for(int i=0; i<size; i++)
         snprintf(str + i*2, 3, "%02x", (unsigned char)digest[i]);
     return str;
@@ -212,12 +230,20 @@ bool hash_init(zckCtx *zck, zckHash *hash, zckHashType *hash_type) {
     if(hash_type->type == ZCK_HASH_SHA1) {
         zck_log(ZCK_LOG_DDEBUG, "Initializing SHA-1 hash");
         hash->ctx = zmalloc(sizeof(SHA_CTX));
+        if (!hash->ctx) {
+           zck_log(ZCK_LOG_ERROR, "OOM in %s", __func__);
+           return false;
+        }
         hash->type = hash_type;
         SHA1_Init((SHA_CTX *) hash->ctx);
         return true;
     } else if(hash_type->type == ZCK_HASH_SHA256) {
         zck_log(ZCK_LOG_DDEBUG, "Initializing SHA-256 hash");
         hash->ctx = zmalloc(sizeof(SHA256_CTX));
+        if (!hash->ctx) {
+           zck_log(ZCK_LOG_ERROR, "OOM in %s", __func__);
+           return false;
+        }
         hash->type = hash_type;
         SHA256_Init((SHA256_CTX *) hash->ctx);
         return true;
@@ -225,6 +251,10 @@ bool hash_init(zckCtx *zck, zckHash *hash, zckHashType *hash_type) {
               hash_type->type <= ZCK_HASH_SHA512_128) {
         zck_log(ZCK_LOG_DDEBUG, "Initializing SHA-512 hash");
         hash->ctx = zmalloc(sizeof(SHA512_CTX));
+        if (!hash->ctx) {
+           zck_log(ZCK_LOG_ERROR, "OOM in %s", __func__);
+           return false;
+        }
         hash->type = hash_type;
         SHA512_Init((SHA512_CTX *) hash->ctx);
         return true;
@@ -240,7 +270,9 @@ bool hash_update(zckCtx *zck, zckHash *hash, const char *message,
         return true;
     if(message == NULL) {
         set_error(zck,
-                  "Hash data is supposed to have %lu bytes, but is NULL", size);
+                  "Hash data is supposed to have %llu bytes, but is NULL",
+                  (long long unsigned) size
+        );
         return false;
     }
     if(size == 0) {
@@ -279,17 +311,29 @@ char *hash_finalize(zckCtx *zck, zckHash *hash) {
     }
     if(hash->type->type == ZCK_HASH_SHA1) {
         unsigned char *digest = zmalloc(SHA1_DIGEST_LENGTH);
+        if (!digest) {
+           zck_log(ZCK_LOG_ERROR, "OOM in %s", __func__);
+           return false;
+        }
         SHA1_Final((sha1_byte*)digest, (SHA_CTX *)hash->ctx);
         hash_close(hash);
         return (char *)digest;
     } else if(hash->type->type == ZCK_HASH_SHA256) {
         unsigned char *digest = zmalloc(SHA256_DIGEST_SIZE);
+        if (!digest) {
+           zck_log(ZCK_LOG_ERROR, "OOM in %s", __func__);
+           return false;
+        }
         SHA256_Final(digest, (SHA256_CTX *)hash->ctx);
         hash_close(hash);
         return (char *)digest;
     } else if(hash->type->type >= ZCK_HASH_SHA512 &&
               hash->type->type <= ZCK_HASH_SHA512_128) {
         unsigned char *digest = zmalloc(SHA512_DIGEST_SIZE);
+        if (!digest) {
+           zck_log(ZCK_LOG_ERROR, "OOM in %s", __func__);
+           return false;
+        }
         SHA512_Final(digest, (SHA512_CTX *)hash->ctx);
         hash_close(hash);
         return (char *)digest;
@@ -381,6 +425,13 @@ int validate_current_chunk(zckCtx *zck) {
 
 int validate_file(zckCtx *zck, zck_log_type bad_checksums) {
     VALIDATE_BOOL(zck);
+    if(zck->has_uncompressed_source) {
+        zck_log(
+            ZCK_LOG_DEBUG,
+            "Skipping full file validation since uncompressed source flag is set"
+        );
+        return 1;
+    }
     char *digest = hash_finalize(zck, &(zck->check_full_hash));
     if(digest == NULL) {
         set_error(zck, "Unable to calculate full file checksum");
@@ -434,10 +485,17 @@ int validate_header(zckCtx *zck) {
     return 1;
 }
 
-/* Returns 1 if data hash matches, -1 if it doesn't and 0 if error */
-int PUBLIC zck_validate_data_checksum(zckCtx *zck) {
+/* Returns 1 if data hash matches, -1 if it doesn't and 0 if error
+ *
+ * For a zchunk file with both compressed and uncompressed checksums, validate
+ * each chunk checksum independently, since there is no data hash */
+int ZCK_PUBLIC_API zck_validate_data_checksum(zckCtx *zck) {
     VALIDATE_READ_BOOL(zck);
 
+    if(zck->has_uncompressed_source) {
+        return validate_checksums(zck, ZCK_LOG_WARNING);
+    }
+
     if(!seek_data(zck, zck->data_offset, SEEK_SET))
         return 0;
     if(!hash_init(zck, &(zck->check_full_hash), &(zck->hash_type)))
@@ -467,7 +525,7 @@ int PUBLIC zck_validate_data_checksum(zckCtx *zck) {
     return ret;
 }
 
-const char PUBLIC *zck_hash_name_from_type(int hash_type) {
+const ZCK_PUBLIC_API char *zck_hash_name_from_type(int hash_type) {
     if(hash_type >= ZCK_HASH_UNKNOWN || hash_type < 0) {
         snprintf(unknown+8, 21, "%i)", hash_type);
         return unknown;
@@ -475,57 +533,64 @@ const char PUBLIC *zck_hash_name_from_type(int hash_type) {
     return HASH_NAME[hash_type];
 }
 
-int PUBLIC zck_get_full_hash_type(zckCtx *zck) {
+int ZCK_PUBLIC_API zck_get_full_hash_type(zckCtx *zck) {
     VALIDATE_INT(zck);
 
     return zck->hash_type.type;
 }
 
-ssize_t PUBLIC zck_get_full_digest_size(zckCtx *zck) {
+ssize_t ZCK_PUBLIC_API zck_get_full_digest_size(zckCtx *zck) {
     VALIDATE_INT(zck);
 
     return zck->hash_type.digest_size;
 }
 
-int PUBLIC zck_get_chunk_hash_type(zckCtx *zck) {
+int ZCK_PUBLIC_API zck_get_chunk_hash_type(zckCtx *zck) {
     VALIDATE_INT(zck);
 
     return zck->index.hash_type;
 }
 
-ssize_t PUBLIC zck_get_chunk_digest_size(zckCtx *zck) {
+ssize_t ZCK_PUBLIC_API zck_get_chunk_digest_size(zckCtx *zck) {
     VALIDATE_INT(zck);
 
     return zck->index.digest_size;
 }
 
-char PUBLIC *zck_get_header_digest(zckCtx *zck) {
+char ZCK_PUBLIC_API *zck_get_header_digest(zckCtx *zck) {
     VALIDATE_PTR(zck);
 
     return get_digest_string(zck->header_digest, zck->hash_type.digest_size);
 }
 
-char PUBLIC *zck_get_data_digest(zckCtx *zck) {
+char ZCK_PUBLIC_API *zck_get_data_digest(zckCtx *zck) {
     VALIDATE_PTR(zck);
 
     return get_digest_string(zck->full_hash_digest, zck->hash_type.digest_size);
 }
 
-char PUBLIC *zck_get_chunk_digest(zckChunk *item) {
+char ZCK_PUBLIC_API *zck_get_chunk_digest(zckChunk *item) {
     if(item == NULL)
         return NULL;
     return get_digest_string(item->digest, item->digest_size);
 }
 
+char ZCK_PUBLIC_API *zck_get_chunk_digest_uncompressed(zckChunk *item) {
+    if(item == NULL || !item->zck->has_uncompressed_source)
+        return NULL;
+    return get_digest_string(item->digest_uncompressed, item->digest_size);
+}
+
+
 /* Returns 1 if all chunks are valid, -1 if even one isn't and 0 if error */
-int PUBLIC zck_find_valid_chunks(zckCtx *zck) {
+int ZCK_PUBLIC_API zck_find_valid_chunks(zckCtx *zck) {
     VALIDATE_READ_BOOL(zck);
 
     return validate_checksums(zck, ZCK_LOG_DEBUG);
 }
 
 /* Returns 1 if all checksums matched, -1 if even one doesn't and 0 if error */
-int PUBLIC zck_validate_checksums(zckCtx *zck) {
+int ZCK_PUBLIC_API zck_validate_checksums(zckCtx *zck) {
     VALIDATE_READ_BOOL(zck);
 
     return validate_checksums(zck, ZCK_LOG_WARNING);
