@@ -44,14 +44,29 @@ static bool check_flags(zckCtx *zck, size_t flags) {
     zck->has_optional_elems = flags & 2;
     if(zck->has_optional_elems)
         flags -= 2;
+    zck->has_uncompressed_source = flags & 4;
+    if(zck->has_uncompressed_source)
+        flags -= 4;
+
     flags = flags & (SIZE_MAX - 1);
     if(flags != 0) {
-        set_fatal_error(zck, "Unknown flags(s) set");
+        set_fatal_error(zck, "Unknown flags(s) set: %i", flags);
         return false;
     }
     return true;
 }
 
+static ssize_t get_flags(zckCtx *zck) {
+    size_t flags = 0;
+    if(zck->has_streams)
+        flags |= 1;
+    if(zck->has_optional_elems)
+        flags |= 2;
+    if(zck->has_uncompressed_source)
+        flags |= 4;
+    return flags;
+}
+
 static bool read_optional_element(zckCtx *zck, size_t id, size_t data_size,
                                   char *data) {
     zck_log(ZCK_LOG_WARNING, "Unknown optional element id %i set", id);
@@ -59,8 +74,18 @@ static bool read_optional_element(zckCtx *zck, size_t id, size_t data_size,
 }
 
 static bool read_header_from_file(zckCtx *zck) {
+    /* Verify that lead_size and header_length have been set */
+    if(zck->lead_size == 0 || zck->header_length == 0) {
+        set_error(zck, "Lead and header sizes are both 0.  Have you run zck_read_lead() yet?");
+        return false;
+    }
+
     /* Allocate header and store any extra bytes at beginning of header */
     zck->header = zrealloc(zck->header, zck->lead_size + zck->header_length);
+    if (!zck->header) {
+        zck_log(ZCK_LOG_ERROR, "OOM in %s", __func__);
+        return false;
+    }
     zck->lead_string = zck->header;
     char *header = zck->header + zck->lead_size;
     size_t loaded = 0;
@@ -73,8 +98,8 @@ static bool read_header_from_file(zckCtx *zck) {
         loaded = zck->header_size - zck->lead_size;
 
     /* Read header from file */
-    zck_log(ZCK_LOG_DEBUG, "Reading the rest of the header: %lu bytes",
-            zck->header_length);
+    zck_log(ZCK_LOG_DEBUG, "Reading the rest of the header: %llu bytes",
+            (long long unsigned) zck->header_length);
     if(loaded < zck->header_length) {
         if(!read_data(zck, header + loaded, zck->header_length - loaded))
             return false;
@@ -83,9 +108,16 @@ static bool read_header_from_file(zckCtx *zck) {
 
     if(!hash_init(zck, &(zck->check_full_hash), &(zck->hash_type)))
         return false;
-    if(!hash_update(zck, &(zck->check_full_hash), zck->header,
-                    zck->hdr_digest_loc))
+    /* If we're reading a detached zchunk header, first five bytes will be
+     * different, breaking the header digest, so let's make things simple
+     * by forcing the first five bytes to be static */
+    if(!hash_update(zck, &(zck->check_full_hash), "\0ZCK1", 5))
         return false;
+    /* Now hash the remaining lead */
+    if(!hash_update(zck, &(zck->check_full_hash), zck->header+5,
+                    zck->hdr_digest_loc-5))
+        return false;
+    /* And the remaining header */
     if(!hash_update(zck, &(zck->check_full_hash), header, zck->header_length))
         return false;
     int ret = validate_header(zck);
@@ -116,6 +148,10 @@ static bool read_preface(zckCtx *zck) {
         return false;
     }
     zck->full_hash_digest = zmalloc(zck->hash_type.digest_size);
+    if (!zck->full_hash_digest) {
+	    zck_log(ZCK_LOG_ERROR, "OOM in %s", __func__);
+	    return false;
+    }
     memcpy(zck->full_hash_digest, header+length, zck->hash_type.digest_size);
     length += zck->hash_type.digest_size;
 
@@ -177,13 +213,13 @@ static bool read_index(zckCtx *zck) {
     }
 
     char *header = NULL;
-    zck_log(ZCK_LOG_DEBUG, "Reading index");
     if(zck->lead_size + zck->preface_size + zck->index_size >
        zck->header_size) {
         set_fatal_error(zck, "Read past end of header");
         return false;
     }
     header = zck->header + zck->lead_size + zck->preface_size;
+    zck_log(ZCK_LOG_DEBUG, "Reading index at 0x%x", (unsigned long)(zck->lead_size + zck->preface_size));
     int max_length = zck->header_size - (zck->lead_size + zck->preface_size);
     if(!index_read(zck, header, zck->index_size, max_length))
         return false;
@@ -221,7 +257,7 @@ static bool read_sig(zckCtx *zck) {
 
     if(zck->header_size >
        zck->lead_size + zck->preface_size + zck->index_size + length)
-        zck_log(ZCK_LOG_WARNING, "There are %lu unused bytes in the header");
+        zck_log(ZCK_LOG_WARNING, "There are unused bytes in the header");
 
     zck->sig_size = length;
     zck->sig_string = header;
@@ -234,6 +270,10 @@ static bool preface_create(zckCtx *zck) {
     int header_malloc = zck->hash_type.digest_size + 4 + 2*MAX_COMP_SIZE;
 
     char *header = zmalloc(header_malloc);
+    if (!header) {
+	    zck_log(ZCK_LOG_ERROR, "OOM in %s", __func__);
+	    return false;
+    }
     size_t length = 0;
 
     /* Write out the full data digest */
@@ -241,10 +281,7 @@ static bool preface_create(zckCtx *zck) {
     length += zck->hash_type.digest_size;
 
     /* Write out flags */
-    size_t flags = 0;
-    if(zck->has_streams)
-        flags &= 1;
-    compint_from_size(header+length, flags, &length);
+    compint_from_size(header+length, get_flags(zck), &length);
 
     /* Write out compression type and index size */
     if(!compint_from_int(zck, header+length, zck->comp.type, &length)) {
@@ -255,10 +292,18 @@ static bool preface_create(zckCtx *zck) {
 
     /* Shrink header to actual size */
     header = zrealloc(header, length);
+    if (!header) {
+        zck_log(ZCK_LOG_ERROR, "OOM in %s", __func__);
+        return false;
+    }
 
     zck->preface_string = header;
     zck->preface_size = length;
-    zck_log(ZCK_LOG_DEBUG, "Generated preface: %lu bytes", zck->preface_size);
+    zck_log(
+        ZCK_LOG_DEBUG,
+        "Generated preface: %llu bytes",
+        (long long unsigned) zck->preface_size
+    );
     return true;
 }
 
@@ -266,6 +311,10 @@ static bool sig_create(zckCtx *zck) {
     char *header = zmalloc(MAX_COMP_SIZE);
     size_t length = 0;
 
+    if (!header) {
+	    zck_log(ZCK_LOG_ERROR, "OOM in %s", __func__);
+	    return false;
+    }
     zck_log(ZCK_LOG_DEBUG, "Calculating %i signatures", zck->sigs.count);
 
     /* Write out signature count and signatures */
@@ -278,7 +327,11 @@ static bool sig_create(zckCtx *zck) {
     }
     zck->sig_string = header;
     zck->sig_size = length;
-    zck_log(ZCK_LOG_DEBUG, "Generated signatures: %lu bytes", zck->sig_size);
+    zck_log(
+        ZCK_LOG_DEBUG,
+        "Generated signatures: %llu bytes",
+        (long long unsigned) zck->sig_size
+    );
     return true;
 }
 
@@ -289,6 +342,10 @@ static bool lead_create(zckCtx *zck) {
     memcpy(header, "\0ZCK1", 5);
     length += 5;
 
+    if (!header) {
+	    zck_log(ZCK_LOG_ERROR, "OOM in %s", __func__);
+	    return false;
+    }
     /* Write out full data and header hash type */
     compint_from_size(header + length, zck->hash_type.type, &length);
     /* Write out header length */
@@ -299,10 +356,18 @@ static bool lead_create(zckCtx *zck) {
     length += zck->hash_type.digest_size;
 
     header = zrealloc(header, length);
+    if (!header) {
+        zck_log(ZCK_LOG_ERROR, "OOM in %s", __func__);
+        return false;
+    }
 
     zck->lead_string = header;
     zck->lead_size = length;
-    zck_log(ZCK_LOG_DEBUG, "Generated lead: %lu bytes", zck->lead_size);
+    zck_log(
+        ZCK_LOG_DEBUG,
+        "Generated lead: %llu bytes",
+        (long long unsigned) zck->lead_size
+    );
     return true;
 }
 
@@ -336,9 +401,16 @@ bool header_create(zckCtx *zck) {
                        zck->index_size + zck->sig_size;
 
     /* Merge everything into one large string */
-    zck_log(ZCK_LOG_DEBUG, "Merging into header: %lu bytes",
-            zck->data_offset);
+    zck_log(
+        ZCK_LOG_DEBUG,
+        "Merging into header: %llu bytes",
+        (long long unsigned) zck->data_offset
+    );
     zck->header = zmalloc(zck->data_offset);
+    if (!zck->header) {
+	    zck_log(ZCK_LOG_ERROR, "OOM in %s", __func__);
+	    return false;
+    }
     size_t offs = 0;
     memcpy(zck->header + offs, zck->lead_string, zck->lead_size);
     free(zck->lead_string);
@@ -385,8 +457,11 @@ bool header_create(zckCtx *zck) {
 bool write_header(zckCtx *zck) {
     VALIDATE_WRITE_BOOL(zck);
 
-    zck_log(ZCK_LOG_DEBUG, "Writing header: %lu bytes",
-            zck->lead_size);
+    zck_log(
+        ZCK_LOG_DEBUG,
+        "Writing header: %llu bytes",
+        (long long unsigned) zck->lead_size
+    );
     if(!write_data(zck, zck->fd, zck->header, zck->header_size))
         return false;
     return true;
@@ -398,6 +473,10 @@ static bool read_lead(zckCtx *zck) {
     int lead = 5 + 2*MAX_COMP_SIZE;
 
     char *header = zmalloc(lead);
+    if (!header) {
+	    zck_log(ZCK_LOG_ERROR, "OOM in %s", __func__);
+	    return false;
+    }
     size_t length = 0;
 
     if(read_data(zck, header, lead) < lead) {
@@ -406,7 +485,9 @@ static bool read_lead(zckCtx *zck) {
         return false;
     }
 
-    if(memcmp(header, "\0ZCK1", 5) != 0) {
+    if(memcmp(header, "\0ZHR1", 5) == 0) {
+        zck->header_only = true;
+    } else if(memcmp(header, "\0ZCK1", 5) != 0) {
         free(header);
         set_error(zck, "Invalid lead, perhaps this is not a zck file?");
         return false;
@@ -447,6 +528,10 @@ static bool read_lead(zckCtx *zck) {
     /* Read header digest */
     zck_log(ZCK_LOG_DEBUG, "Reading header digest");
     header = zrealloc(header, length + zck->hash_type.digest_size);
+    if (!header) {
+        zck_log(ZCK_LOG_ERROR, "OOM in %s", __func__);
+        return false;
+    }
     size_t to_read = 0;
     if(lead < length + zck->hash_type.digest_size)
         to_read = length + zck->hash_type.digest_size - lead;
@@ -475,6 +560,11 @@ static bool read_lead(zckCtx *zck) {
         return false;
     }
     zck->header_digest = zmalloc(zck->hash_type.digest_size);
+    if (!zck->header_digest) {
+	    zck_log(ZCK_LOG_ERROR, "OOM in %s", __func__);
+	    free(header);
+	    return false;
+    }
     memcpy(zck->header_digest, header + length, zck->hash_type.digest_size);
     length += zck->hash_type.digest_size;
 
@@ -487,10 +577,12 @@ static bool read_lead(zckCtx *zck) {
         hash_reset(&(zck->hash_type));
         free(zck->header_digest);
         zck->header_digest = NULL;
-        set_error(zck,
-                  "Header length (%lu) doesn't match requested header length "
-                  "(%lu)", zck->header_length + length,
-                  zck->prep_hdr_size);
+        set_error(
+            zck,
+            "Header length (%llu) doesn't match requested header length (%llu)",
+            (long long unsigned) zck->header_length + length,
+            (long long unsigned) zck->prep_hdr_size
+        );
         return false;
     }
     /* Store pre-header */
@@ -498,17 +590,21 @@ static bool read_lead(zckCtx *zck) {
     zck->header_size = lead;
     zck->lead_string = header;
     zck->lead_size = length;
-    zck_log(ZCK_LOG_DEBUG, "Parsed lead: %lu bytes", length);
+    zck_log(
+        ZCK_LOG_DEBUG,
+        "Parsed lead: %llu bytes",
+        (long long unsigned) length
+    );
     return true;
 }
 
-bool PUBLIC zck_read_lead(zckCtx *zck) {
+bool ZCK_PUBLIC_API zck_read_lead(zckCtx *zck) {
     VALIDATE_BOOL(zck);
 
     return read_lead(zck);
 }
 
-bool PUBLIC zck_validate_lead(zckCtx *zck) {
+bool ZCK_PUBLIC_API zck_validate_lead(zckCtx *zck) {
     VALIDATE_BOOL(zck);
 
     int retval = read_lead(zck);
@@ -530,7 +626,7 @@ bool PUBLIC zck_validate_lead(zckCtx *zck) {
     return retval;
 }
 
-bool PUBLIC zck_read_header(zckCtx *zck) {
+bool ZCK_PUBLIC_API zck_read_header(zckCtx *zck) {
     VALIDATE_READ_BOOL(zck);
 
     if(!read_header_from_file(zck))
@@ -544,17 +640,17 @@ bool PUBLIC zck_read_header(zckCtx *zck) {
     return true;
 }
 
-ssize_t PUBLIC zck_get_header_length(zckCtx *zck) {
+ssize_t ZCK_PUBLIC_API zck_get_header_length(zckCtx *zck) {
     VALIDATE_INT(zck);
     return zck->lead_size + zck->header_length;
 }
 
-ssize_t PUBLIC zck_get_lead_length(zckCtx *zck) {
+ssize_t ZCK_PUBLIC_API zck_get_lead_length(zckCtx *zck) {
     VALIDATE_INT(zck);
     return zck->lead_size;
 }
 
-ssize_t PUBLIC zck_get_data_length(zckCtx *zck) {
+ssize_t ZCK_PUBLIC_API zck_get_data_length(zckCtx *zck) {
     VALIDATE_INT(zck);
     zckChunk *idx = zck->index.first;
     while(idx->next != NULL)
@@ -562,7 +658,17 @@ ssize_t PUBLIC zck_get_data_length(zckCtx *zck) {
     return idx->start + idx->comp_length;
 }
 
-ssize_t PUBLIC zck_get_length(zckCtx *zck) {
+ssize_t ZCK_PUBLIC_API zck_get_length(zckCtx *zck) {
     VALIDATE_INT(zck);
     return zck_get_header_length(zck) + zck_get_data_length(zck);
 }
+
+ssize_t ZCK_PUBLIC_API zck_get_flags(zckCtx *zck) {
+    VALIDATE_INT(zck);
+    return get_flags(zck);
+}
+
+bool ZCK_PUBLIC_API zck_is_detached_header(zckCtx *zck) {
+    VALIDATE_BOOL(zck);
+    return zck->header_only;
+}
